import os
# 2023-05-13 import oschmod
import gzip
import csv
import json
from pathlib import Path
from qualys_etl.etld_lib import etld_lib_config
from qualys_etl.etld_lib import etld_lib_sqlite_tables
from qualys_etl.etld_lib import etld_lib_functions
from qualys_etl.etld_lib import etld_lib_oschmod as oschmod
from qualys_etl.etld_lib import etld_lib_log_validation


class CsvDistribution():

    def __init__(self,
                 target_csv_path,
                 target_csv_data_directory
                 ):

        self.csv_writer = None
        self.target_csv_path = target_csv_path
        self.target_csv_data_directory = target_csv_data_directory
        if not target_csv_data_directory.is_dir():
            self.create_directory(target_csv_data_directory)

    def set_csv_writer_with_file_handle(self, file_handle):
        csv_key = etld_lib_config.get_python_csv_quoting_option_key(
            csv_value=etld_lib_config.csv_distribution_python_csv_quoting
        )
        etld_lib_functions.logger.info(f"python_csv_options: {csv_key} {etld_lib_config.get_csv_distribution_var_dict()}")
        #etld_lib_functions.logger.info(
        #    f"tested_mysql_load_example: {etld_lib_config.csv_distribution_mysql_load_example} "
        #    f"- note bash requires additional \\")
        csv_writer = csv.writer(file_handle,
                                quoting=etld_lib_config.csv_distribution_python_csv_quoting,
                                delimiter=etld_lib_config.csv_distribution_python_csv_dialect_delimiter,
                                doublequote=etld_lib_config.csv_distribution_python_csv_dialect_doublequote,
                                escapechar=etld_lib_config.csv_distribution_python_csv_dialect_escapechar,
                                lineterminator=etld_lib_config.csv_distribution_python_csv_dialect_lineterminator,
                                quotechar=etld_lib_config.csv_distribution_python_csv_dialect_quotechar,
                                skipinitialspace=etld_lib_config.csv_distribution_python_csv_dialect_skipinitialspace,
                                strict=etld_lib_config.csv_distribution_python_csv_dialect_strict
                                )
        self.csv_writer = csv_writer

    def write_csv_row_to_file_handle(self, row):
        self.csv_writer.writerow(row)

    @staticmethod
    def create_directory(dir_path=None):
        try:
            if dir_path is not None:
                os.makedirs(dir_path, exist_ok=True)
                # 2023-05-13 oschmod.set_mode(dir_path, "a+rwx,g-rwx,o-rwx")
                oschmod.set_mode(dir_path, "u+rwx,g-rwx,o-rwx")
        except Exception as e:
            raise Exception(f"Could not create directory {dir_path}")

    @staticmethod
    def get_csv_writer(csv_file, csv_open_method=open):
        #delimiter=',',
        #lineterminator='\n',
        #quoting=csv.QUOTE_NONE,
        #escapechar='\\'
        csv_key = etld_lib_config.get_python_csv_quoting_option_key(
            csv_value=etld_lib_config.csv_distribution_python_csv_quoting
        )
        etld_lib_functions.logger.info(f"python_csv_options: {csv_key} {etld_lib_config.get_csv_distribution_var_dict()}")
        #etld_lib_functions.logger.info(
        #    f"tested_mysql_load_example: {etld_lib_config.csv_distribution_mysql_load_example} "
        #    f"- note bash requires additional \\")
        csv_writer = csv.writer(csv_open_method(csv_file, 'wt', newline=''),
                                quoting=etld_lib_config.csv_distribution_python_csv_quoting,
                                delimiter=etld_lib_config.csv_distribution_python_csv_dialect_delimiter,
                                doublequote=etld_lib_config.csv_distribution_python_csv_dialect_doublequote,
                                escapechar=etld_lib_config.csv_distribution_python_csv_dialect_escapechar,
                                lineterminator=etld_lib_config.csv_distribution_python_csv_dialect_lineterminator,
                                quotechar=etld_lib_config.csv_distribution_python_csv_dialect_quotechar,
                                skipinitialspace=etld_lib_config.csv_distribution_python_csv_dialect_skipinitialspace,
                                strict=etld_lib_config.csv_distribution_python_csv_dialect_strict
                                )
        return csv_writer

    @staticmethod
    def write_csv_row(csv_writer, row):
        csv_writer.writerow(row)

    @staticmethod
    def prepare_csv_row(headers, row, pragma, csv_max_field_size) -> list:
        row_fields = []
        for idx, fieldname in enumerate(headers):
            field_data = row[idx]
            if pragma[idx][2] == 'INTEGER':
                field_data = str(row[idx]).strip()
                if not str(row[idx]).isnumeric():
                    field_data = '0'
            elif str(field_data).__contains__('\r'):
                    field_data = str(field_data).replace('\r', '')
            if 'dns' in str(fieldname).lower():
                field_data = str(field_data).replace('\r', '')
                field_data = str(field_data).replace('\n', '')
            row_fields.append(f"{str(field_data)[0:int(csv_max_field_size)]}")
        return row_fields


def get_target_temp_csv_file_name(batch_date, batch_number, table_name, csv_file_open_method=open):
    batch_date = str(batch_date).replace(' ', 'T') + 'Z'
    table_name_utc_run_datetime = f"_utc_run_datetime_{batch_date}"
    table_name_batch_number = '_batch_000000'
    if '_batch_' in batch_number:
        table_name_batch_number = f'{batch_number}'
    elif batch_number.startswith('batch_'):
        table_name_batch_number = f'_{batch_number}'
    elif 'batch' not in batch_number:
        table_name_batch_number = f'_batch_{batch_number}'

    # 'Q_Host_List_Detection_HOSTS_utc_run_datetime_2022-11-16T17:31:11Z_batch_000001'
    target_csv_file_name = f"{table_name}{table_name_utc_run_datetime}{table_name_batch_number}"
    if csv_file_open_method == open:
        target_csv_file_name = f"{target_csv_file_name}.csv.tmp"
    elif csv_file_open_method == gzip.open:
        target_csv_file_name = f"{target_csv_file_name}.csv.gz.tmp"
    else:
        etld_lib_functions.logger.error(f"ERROR: Unknown csv_file_open_method: {csv_file_open_method}")
        etld_lib_functions.logger.error(f"ERROR: table_name: {table_name}, batch_date: {batch_date}, batch_number: {batch_number}")
        exit(1)

    return target_csv_file_name, table_name_utc_run_datetime, table_name_batch_number


def get_target_csv_file_name(batch_date, batch_number, table_name, csv_file_open_method=open):
    batch_date = str(batch_date).replace(' ', 'T') + 'Z'
    table_name_utc_run_datetime = f"_utc_run_datetime_{batch_date}"
    table_name_batch_number = '_batch_000000'
    if '_batch_' in batch_number:
        table_name_batch_number = f'{batch_number}'
    elif batch_number.startswith('batch_'):
        table_name_batch_number = f'_{batch_number}'
    elif 'batch' not in batch_number:
        table_name_batch_number = f'_batch_{batch_number}'

    # 'Q_Host_List_Detection_HOSTS_utc_run_datetime_2022-11-16T17:31:11Z_batch_000001'
    target_csv_file_name = f"{table_name}{table_name_utc_run_datetime}{table_name_batch_number}"
    if csv_file_open_method == open:
        target_csv_file_name = f"{target_csv_file_name}.csv"
    elif csv_file_open_method == gzip.open:
        target_csv_file_name = f"{target_csv_file_name}.csv.gz"
    else:
        etld_lib_functions.logger.error(f"ERROR: Unknown csv_file_open_method: {csv_file_open_method}")
        etld_lib_functions.logger.error(f"ERROR: table_name: {table_name}, batch_date: {batch_date}, batch_number: {batch_number}")
        exit(1)

    target_tmp_csv_file_name = target_csv_file_name + ".tmp"
    return target_csv_file_name, table_name_utc_run_datetime, table_name_batch_number, target_tmp_csv_file_name


def get_table_name_utc_run_datetime(sqlite_obj, etl_workflow_location_dict):
    sqlite_obj.reopen_connection_and_cursor_as_sqlite3_row_factory()
    database_table_names_list = sqlite_obj.get_all_table_names_from_database()
    table_name_utc_run_datetime = f"_utc_run_datetime_1970-01-01T00:00:00Z"
    table_name_batch_number = '_batch_000000'
    # CREATE FILENAME Q_TABLENAME_utc_run_datetime_2022-11-17T16:45:37Z_batch_000001.csv.gz
    for database_table_name in database_table_names_list:
        if str(database_table_name).endswith("_Status"):
            sqlite_obj.reopen_connection_and_cursor_as_sqlite3_row_factory()
            sqlite_obj.cursor.execute(f"select * from {database_table_name}")
            row = sqlite_obj.cursor.fetchone()
            if row is None:
                etld_lib_functions.logger.error(f"No Status Rows Found in Database: {etl_workflow_location_dict['sqlite_file']}")
                etld_lib_functions.logger.error(f"Please rerun to rebuild database: {etl_workflow_location_dict['sqlite_file']}")
                exit(1)
            if 'STATUS_DETAIL' in row.keys():
                json_string = row['STATUS_DETAIL']
                status_detail = json.loads(json_string)
                batch_date = str(status_detail['BATCH_DATE']).replace(' ', 'T') + 'Z'
                batch_number = 'batch_000000'
                dummy_value, \
                table_name_utc_run_datetime, \
                table_name_batch_number, tmp_csv_file_name = \
                    get_target_csv_file_name(batch_date=batch_date,
                                             batch_number=batch_number,
                                             table_name='Q_DUMMY',
                                             )
    return table_name_utc_run_datetime, table_name_batch_number


def get_csv_info_dict(batch_date, batch_number, table_name, csv_file_open_method=open, distribution_dir=None):
    csv_info_dict = {}
    batch_date = str(batch_date).replace(' ', 'T') + 'Z'
    table_name_utc_run_datetime = f"_utc_run_datetime_{batch_date}"
    table_name_batch_number = '_batch_000000'
    if '_batch_' in batch_number:
        table_name_batch_number = f'{batch_number}'
    elif batch_number.startswith('batch_'):
        table_name_batch_number = f'_{batch_number}'
    elif 'batch' not in batch_number:
        table_name_batch_number = f'_batch_{batch_number}'

    # 'Q_Host_List_Detection_HOSTS_utc_run_datetime_2022-11-16T17:31:11Z_batch_000001'
    target_csv_file_name = f"{table_name}{table_name_utc_run_datetime}{table_name_batch_number}"
    target_tmp_csv_file_name = ""
    if csv_file_open_method == open:
        target_csv_file_name = f"{target_csv_file_name}.csv"
        target_tmp_csv_file_name = target_csv_file_name + ".tmp"
    elif csv_file_open_method == gzip.open:
        target_csv_file_name = f"{target_csv_file_name}.csv.gz"
        target_tmp_csv_file_name = target_csv_file_name + ".tmp"
    else:
        etld_lib_functions.logger.error(f"ERROR: Unknown csv_file_open_method: {csv_file_open_method}")
        etld_lib_functions.logger.error(
            f"ERROR: table_name: {table_name}, batch_date: {batch_date}, batch_number: {batch_number}")
        exit(1)

    csv_info_dict['compression_method'] = csv_file_open_method
    csv_info_dict['target_csv_file_name'] = target_csv_file_name
    csv_info_dict['table_name_utc_run_datetime'] = table_name_utc_run_datetime
    csv_info_dict['table_name_batch_number'] = table_name_batch_number
    csv_info_dict['target_tmp_csv_file_name'] = target_tmp_csv_file_name
    csv_info_dict['distribution_dir'] = distribution_dir
    # FINAL CSV FILENAME Target
    csv_info_dict['target_csv_file_path'] = Path(distribution_dir, target_csv_file_name)
    csv_info_dict['target_tmp_csv_file_path'] = Path(distribution_dir, target_tmp_csv_file_name)

    csv_info_dict['csv_obj'] = \
        CsvDistribution(target_csv_path=csv_info_dict['target_tmp_csv_file_path'],
                        target_csv_data_directory=Path(csv_info_dict['distribution_dir']))

    return csv_info_dict


def get_one_etl_workflow_csv_data(etl_workflow, counter_obj_dict, exclude_table_names=None):

    if exclude_table_names is None:
        exclude_table_names = []

    final_exit_status = False
    target_csv_path_list = []
    #passed_all_tests = True
    #if etld_lib_config.test_system_do_not_test_intermediary_extracts_flag:
    #    passed_all_tests = True  # Skip during etl_test_system
    #else:
    #    passed_all_tests = \
    #        etld_lib_log_validation.validate_log_has_no_errors_prior_to_distribution(etl_workflow)

    etl_workflow_location_dict = \
        etld_lib_config.get_etl_workflow_data_location_dict(etl_workflow)

    #if passed_all_tests and len(etl_workflow_location_dict) > 0:
    if len(etl_workflow_location_dict) > 0:
        sqlite_obj = etld_lib_sqlite_tables.SqliteObj(sqlite_file=etl_workflow_location_dict['sqlite_file'])
        sqlite_obj.open_connection_and_cursor()

        all_database_table_names_list = sqlite_obj.get_all_table_names_from_database()
        database_table_names_list = []
        for database_table_name in all_database_table_names_list:
            if database_table_name in exclude_table_names:
                pass
            else:
                database_table_names_list.append(database_table_name)

        table_name_utc_run_datetime, table_name_batch_number = \
            get_table_name_utc_run_datetime(sqlite_obj, etl_workflow_location_dict)

        for database_table_name in database_table_names_list:
            target_csv_file_name = f"{database_table_name}{table_name_utc_run_datetime}{table_name_batch_number}"
            target_csv_path = Path(etl_workflow_location_dict['distribution_dir'], f"{target_csv_file_name}.csv")
            if etl_workflow_location_dict['open_file_compression_method'] == open:
                pass
            elif etl_workflow_location_dict['open_file_compression_method'] == gzip.open:
                target_csv_path = Path(f"{str(target_csv_path)}.gz.tmp")
            target_csv_path_list.append(target_csv_path)

            sqlite_obj.reopen_connection_and_cursor_as_sqlite3_row_factory()
            pragma = sqlite_obj.get_pragma(sqlite_obj.cursor, database_table_name)
            sqlite_obj.reopen_connection_and_cursor_as_sqlite3_row_factory()
            sqlite_obj.cursor.execute(f"select * from {database_table_name}")
            headers = prepare_headers_from_sqlite(cursor=sqlite_obj.cursor)


            csv_obj = CsvDistribution(
                target_csv_path=target_csv_path,
                target_csv_data_directory=etl_workflow_location_dict['distribution_dir']
            )
            csv_writer = csv_obj.get_csv_writer(
                csv_file=target_csv_path,
                csv_open_method=etl_workflow_location_dict['open_file_compression_method']
            )


            counter_obj_dict[database_table_name] = \
                etld_lib_functions.DisplayCounterToLog(
                    display_counter_at=etl_workflow_location_dict['csv_distribution_display_counter_at'],
                    logger_func=etld_lib_functions.logger.info,
                    display_counter_log_message=f"rows added to csv file from "
                                                f"table {database_table_name}")

            while True:
                row = sqlite_obj.cursor.fetchone()
                if row is None:
                    target_path = str(target_csv_path).replace('.tmp', '')
                    target_csv_path.rename(target_path)
                    break
                new_row = []
                for field in row:
                    new_row.append(field)
                row = csv_obj.prepare_csv_row(headers=headers,
                                              row=new_row,
                                              pragma=pragma,
                                              csv_max_field_size=etl_workflow_location_dict['csv_max_field_size'])
                csv_obj.write_csv_row(csv_writer, row)
                counter_obj_dict[database_table_name].update_counter_and_display_to_log()

        final_exit_status = True

        for database_table_name in database_table_names_list:
            counter_obj_dict[database_table_name].display_final_counter_to_log()

        log_csv_location_information(target_csv_path_list)

    return final_exit_status


def log_csv_location_information(target_csv_path_list: list):
    for target_csv_path in target_csv_path_list:
        etld_lib_functions.logger.info(f"File ready for distribution: {str(target_csv_path).replace('.tmp','')}")


def distribute_csv_data_for_one_workflow(distribution_csv_flag=False,
                                         etl_workflow='knowledgebase_etl_workflow',
                                         distribution_csv_flag_name='kb_distribution_csv_flag',
                                         exclude_table_names=None
                                         ):
    if exclude_table_names == None:
        exclude_table_names = []


    counter_obj_dict = {}
    if distribution_csv_flag:
        try:
            etl_workflow_location_dict = \
                etld_lib_config.get_etl_workflow_data_location_dict(etl_workflow)
            if len(etl_workflow_location_dict) > 0:
                get_one_etl_workflow_csv_data(
                    etl_workflow=etl_workflow,
                    counter_obj_dict=counter_obj_dict,
                    exclude_table_names=exclude_table_names)
            else:
                raise Exception(f"Could not retrieve etl_workflow csv data for: {etl_workflow}")
        except Exception as e:
            etld_lib_functions.logger.error(f"Distribution Program aborted, Exception: {e}")
            exit(1)
    else:
        etld_lib_functions.logger.info(f"No distribute_csv_data_for_one_workflow. {distribution_csv_flag_name} set to: {distribution_csv_flag} ")


def prepare_headers_from_sqlite(cursor):
    headers = [i[0] for i in cursor.description]
    return headers


def get_all_etl_workflow_csv_data():
    csv_obj_list = []
    counter_obj_dict = {}
    final_exit_status = 0
    for etl_workflow in etld_lib_config.etl_workflow_list:
        if get_one_etl_workflow_csv_data(
                etl_workflow=etl_workflow,
                counter_obj_dict=counter_obj_dict,
                exclude_table_names=None
        ):
            pass
        else:
            final_exit_status = 2

        if final_exit_status == 0:
            pass
        else:
            etld_lib_functions.logger.error(f"ERROR found in csv creation, exit status is {final_exit_status}")

    for table_name in counter_obj_dict:
        counter_obj_dict[table_name].display_final_counter_to_log()

    exit(final_exit_status)


if __name__ == "__main__":
    etld_lib_functions.main(my_logger_prog_name='etld_lib_prepare_distribution')
    etld_lib_config.main()
    get_all_etl_workflow_csv_data()


import sqlite3
import re
import csv
import sys
import json
from pathlib import Path
from qualys_etl.etld_lib import etld_lib_functions
from qualys_etl.etld_lib import etld_lib_datetime
from qualys_etl.etld_lib import etld_lib_config
global count_rows_added_to_table


# def get_q_knowledgebase_min_max_dates(etl_workflow_sqlite_file_path):
#     min_date = None
#     max_date = None
#     try:
#         conn = sqlite3.connect(etl_workflow_sqlite_file_path)
#         cursor = conn.cursor()
#         select_statement = 'select min(LAST_SERVICE_MODIFICATION_DATETIME) MIN_DATE, ' \
#                            'max(LAST_SERVICE_MODIFICATION_DATETIME) MAX_DATE ' \
#                            'from Q_KnowledgeBase;'
#         cursor.execute(select_statement)
#         records = cursor.fetchall()
#         for row in records:
#             min_date = row[0]
#             max_date = row[1]
#         conn.close()
#     except Exception as e:
#         min_date = 'Not Found'
#         max_date = 'Not Found'
#
#     return min_date, max_date


# def get_q_table_min_max_dates(etl_workflow_sqlite_file_path, datetime_field_name, table_name):
#     min_date = None
#     max_date = None
#     try:
#         conn = sqlite3.connect(etl_workflow_sqlite_file_path)
#         cursor = conn.cursor()
#         select_statement = f'select min({datetime_field_name}) MIN_DATE, ' \
#                            f'max({datetime_field_name}) MAX_DATE ' \
#                            f'from {table_name};'
#         cursor.execute(select_statement)
#         records = cursor.fetchall()
#         for row in records:
#             min_date = row[0]
#             max_date = row[1]
#         conn.close()
#     except Exception as e:
#         min_date = 'Not Found'
#         max_date = 'Not Found'
#
#     return min_date, max_date


# def create_table_q_knowledgebase_in_host_list_detection():
#
#     table_name = 'Q_KnowledgeBase_In_Host_List_Detection'
#     kb_table_name = 'Q_KnowledgeBase'
#     table_statement = f'CREATE TABLE {table_name} as ' \
#                       f'select * from {kb_table_name} ' \
#                       f'where QID in (select distinct(QID) from Q_Host_List_Detection);'
#     try:
#
#         conn = sqlite3.connect()
#         conn.execute(f"DROP TABLE IF EXISTS {table_name};")
#         conn.execute(table_statement)
#         conn.execute(f"DROP TABLE IF EXISTS {kb_table_name};")
#         conn.commit()
#         conn.close()
#         conn = sqlite3.connect(etl_workflow_sqlite_file_path)
#         conn.isolation_level = None
#         conn.execute("VACUUM")
#         conn.close()
#     except Exception as e:
#         etld_lib_functions.logger.error(f"Error creating table: {table_name}, please retry after fixing error")
#         etld_lib_functions.logger.error(f"Exception: {e}")
#         exit(1)


class SqliteObj:
    def __init__(self, sqlite_file, read_only=False):
        self.sqlite_file = str(sqlite_file)
        self.read_only = read_only
        self.connection = None
        self.cursor = None
        self.isolation_level = None
        self.count_rows_added_to_table = 0
        self.status_table_recreated = False
        self.open_connection_and_cursor(read_only=read_only)

    def get_pragma(self, cursor, table_name):
        # pragma[0]['name'] - field 0 name 'field name in table'
        # pragma[0]['type'] - field 0 type 'INTEGER'
        sqlPragma = f'PRAGMA table_info({table_name})'
        try:
            cursor.execute(sqlPragma)
            result = cursor.fetchall()
        except sqlite3.DatabaseError as e:
            raise Exception(f"Exception: {e} - Could not obtain pragma for table: {table_name}.")
        return result

    def get_all_table_names_from_database(self) -> list:
        database_table_names = []
        try:
            self.cursor.execute("select tbl_name from sqlite_master where type == 'table'")
            for create_table_list in self.cursor.fetchall():
                create_table_statement = str(create_table_list[0])
                database_table_names.append(create_table_statement)
        except sqlite3.DatabaseError as e:
            raise Exception(f"Exception: {e} - Could not obtain table names for {self.sqlite_file}.")
        return database_table_names

    def get_create_table_statement_from_database(self, table_name) -> str:
        create_table_statement = ""
        try:
            select_statement = "select sql from sqlite_master " \
                               "where type == 'table' and " \
                               f"tbl_name == '{table_name}'"
            self.cursor.execute(select_statement)
            create_table_statement = self.cursor.fetchone()
            create_table_statement = str(create_table_statement[0])
        except sqlite3.DatabaseError as e:
            raise Exception(f"Exception: {e} - Could not obtain table names for {self.sqlite_file}.")
        return create_table_statement

    def reopen_connection_and_cursor_as_sqlite3_row_factory(self):
        try:
           self.close_connection()
        except Exception as e:
            pass # Ignore errors

        try:
            self.connection = sqlite3.connect(database=self.sqlite_file, timeout=300)
            self.connection.row_factory = sqlite3.Row
            self.cursor = self.connection.cursor()
        except Exception as e:
            etld_lib_functions.logger.error(f"Error getting row factory connnection cursor obj: {self.sqlite_file}, please retry after fixing error")
            etld_lib_functions.logger.error(f"Exception: {e}")
            exit(1)

    def open_connection_and_cursor(self, read_only=False, open_cursor=True, enable_row_factory=True):
        try:
            if read_only is True and Path(self.sqlite_file).exists:
                self.connection = sqlite3.connect(f'file:{str(self.sqlite_file)}?mode=ro', uri=True, timeout=300)
            else:
                self.connection = sqlite3.connect(database=self.sqlite_file, timeout=300)
            if enable_row_factory is True:
                self.connection.row_factory = sqlite3.Row
            if open_cursor is True:
                self.cursor = self.connection.cursor()

        except Exception as e:
            etld_lib_functions.logger.error(f"Error getting cursor obj: {self.sqlite_file}, please retry after fixing error")
            etld_lib_functions.logger.error(f"Exception: {e}")
            exit(1)

    def attach_database_to_connection(self, database_as_name, database_sqlite_file: str, mode='ro'):
        attach_database_sql = f'ATTACH DATABASE ? AS {database_as_name}'
        # unsupported on redhat 8.4 self.connection.execute(attach_database_sql, (f'file:{str(database_sqlite_file)}?mode={mode}',))
        try:
            self.connection.execute(attach_database_sql, (str(database_sqlite_file),))
        except sqlite3.Error as err:
            etld_lib_functions.logger.error(f"sqlite3.Error attaching database: {self.sqlite_file}, please retry after fixing error")
            etld_lib_functions.logger.error(f"Exception: {err}")
            exit(1)
        except Exception as e:
            etld_lib_functions.logger.error(f"Error attaching database: {self.sqlite_file}, please retry after fixing error")
            etld_lib_functions.logger.error(f"Exception: {e}")
            exit(1)

    def detach_database_to_connection(self, database_as_name, database_sqlite_file: str, mode='ro'):
        detach_database_sql = f'DETACH DATABASE {database_as_name}'
        # unsupported on redhat 8.4 self.connection.execute(attach_database_sql, (f'file:{str(database_sqlite_file)}?mode={mode}',))
        try:
            self.connection.execute(detach_database_sql,)
        except sqlite3.Error as err:
            etld_lib_functions.logger.error(f"sqlite3.Error detaching database {database_as_name} from {self.sqlite_file}, please retry after fixing error")
            etld_lib_functions.logger.error(f"Exception: {err}")
            exit(1)
        except Exception as e:
            etld_lib_functions.logger.error(f"Error outside of sqlite3 detaching database {database_as_name}, {self.sqlite_file}, please retry after fixing error")
            etld_lib_functions.logger.error(f"Exception: {e}")
            exit(1)

    def get_table_columns(self, table_name) -> list:
        try:
            sql = f"select * from {table_name} where 1=0;"
            self.cursor.execute(sql)
            return [d[0] for d in self.cursor.description]
        except Exception as e:
            etld_lib_functions.logger.info(f"Table not found: {table_name}")
            etld_lib_functions.logger.info(f"Exception: {e}")
            return []

    def get_min_max_dates_from_table_column(self, table_name, column_name):
        select_statement = f'select min({column_name}) MIN_DATE, ' \
                           f'max({column_name}) MAX_DATE ' \
                           f'from {table_name};'
        self.cursor.execute(select_statement)
        records = self.cursor.fetchall()
        min_date=""
        max_date=""
        if len(records) == 1:
            min_date = records[0][0]
            max_date = records[0][1]
        return {'min_date': min_date, 'max_date': max_date}

    def get_distinct_batch_numbers_from_table_column(self, table_name, column_name='BATCH_NUMBER') -> list:
        select_statement = f'select distinct({column_name}) ' \
                           f'from {table_name};'
        self.cursor.execute(select_statement)
        records = self.cursor.fetchall()
        record_list = []
        for record in records:
            record_list.append(record)

        return record_list

    def select_rows_from_status_table(self, table_name):
        select_statement = f'select * from {table_name} limit 100;'
        self.cursor.execute(select_statement)
        records = self.cursor.fetchall()
        return records

    def execute_statement(self, statement):
        self.cursor.execute(f"{statement};")


    # def create_table(self,
    #                  table_name=None, csv_columns=[],
    #                  key=False, csv_column_types={}):
    #     try:
    #         if key is False:
    #             table_statement = self.create_table_statement_no_primary_key(table_name, csv_columns)
    #         else:
    #             table_statement = self.create_table_statement(key, table_name, csv_columns, csv_column_types)
    #
    #         self.cursor.execute(table_statement)
    #         table_statement_one_line = table_statement.replace("\n", " ")
    #         etld_lib_functions.logger.info(f"sqlite create table statement: {table_statement_one_line}")
    #         self.connection.commit()
    #     except Exception as e:
    #         etld_lib_functions.logger.error(f"Error creating table: {table_name}, please retry after fixing error")
    #         etld_lib_functions.logger.error(f"Exception: {e}")
    #         exit(1)

    def drop_and_recreate_table(self, table_name: str, csv_columns: list, csv_column_types: dict, key=False):
        try:
            self.cursor.execute(f"DROP TABLE IF EXISTS {table_name};")
            if key is False:
                table_statement = \
                    self.create_table_statement_no_primary_key_and_column_types(
                        table_name=table_name,
                        csv_columns=csv_columns,
                        csv_column_types=csv_column_types)
            else:
                table_statement = self.create_table_statement(
                    primary_key=key,
                    table_name=table_name,
                    csv_columns=csv_columns,
                    csv_column_types=csv_column_types)

            self.cursor.execute(table_statement)
            table_statement_one_line = table_statement.replace("\n", " ")
            etld_lib_functions.logger.info(f"sqlite create table statement: {table_statement_one_line}")
            self.connection.commit()
        except Exception as e:
            etld_lib_functions.logger.error(f"Error creating table: {table_name}, please retry after fixing error")
            etld_lib_functions.logger.error(f"Exception: {e}")
            exit(1)

    @staticmethod
    def create_table_statement(primary_key, table_name, csv_columns, csv_column_types={}):
        table_statement = f"CREATE TABLE IF NOT EXISTS {table_name} (\n"
        primary_keys = ""
        if isinstance(primary_key, list):
            pass
        elif primary_key in csv_column_types.keys():
            table_statement = f"{table_statement}{primary_key} {csv_column_types[primary_key]} PRIMARY KEY NOT NULL,\n"
        else:
            table_statement = f"{table_statement}{primary_key} CHAR PRIMARY KEY NOT NULL,\n"

        for column in csv_columns:
            if isinstance(primary_key, list) or primary_key != column:
                if column in csv_column_types.keys():
                    column_type = csv_column_types[column]
                    table_statement = f"{table_statement}{column} {column_type},\n"
                else:
                    table_statement = f"{table_statement}{column} TEXT,\n"

        # table_statement = f"{table_statement}Row_Last_Updated TEXT,\n"

        table_statement = re.sub(",\n$", "", table_statement)

        if isinstance(primary_key, list):
            primary_keys = ", ".join(primary_key)
            table_statement = f"{table_statement},\nPRIMARY KEY ({primary_keys})\n"

        table_statement = table_statement + ");"
        return table_statement

    @staticmethod
    def create_table_statement_no_primary_key_and_column_types(table_name: str, csv_columns: list, csv_column_types: dict):
        table_statement = f"CREATE TABLE IF NOT EXISTS {table_name} (\n"
        for column in csv_columns:
            if column in csv_column_types.keys():
                column_type = csv_column_types[column]
                table_statement = f"{table_statement}{column} {column_type},\n"
            else:
                table_statement = f"{table_statement}{column} TEXT,\n"

        # table_statement = f"{table_statement}Row_Last_Updated TEXT,\n"

        table_statement = re.sub(",\n$", "", table_statement)

        table_statement = table_statement + ");"
        return table_statement

    # @staticmethod
    # def create_table_statement_no_primary_key(table_name, csv_columns):
    #     table_statement_list = []
    #     table_statement_list.append(f"CREATE TABLE IF NOT EXISTS {table_name} (\n")
    #     for column in csv_columns:
    #         table_statement_list.append(f"{column} TEXT,\n")
    #
    #     table_statement_list[-1] = re.sub(",\n$", "", table_statement_list[-1])
    #     # table_statement_list.append(f"Row_Last_Updated TEXT")
    #
    #     table_statement_list.append(");")
    #     table_statement = "".join(table_statement_list)
    #     return table_statement

    # def get_csv_column_type(self, field_name: str, csv_column_types):
    #     if field_name in csv_column_types:
    #         column_type = csv_column_types()[field_name]
    #     else:
    #         column_type = ''
    #     return column_type

    def prepare_database_row_vmpc(self, item_dict: dict, csv_columns: list,
                                  csv_column_types: list, batch_date: str, batch_number: str):
        row_in_sqlite_form = []
        for field_name in csv_columns:  # Iterate through expected columns (contract)
            if field_name in item_dict.keys():  # Iterate through columns found in dictionary
                item_dict[field_name] = \
                    self.prepare_database_field_vmpc(item_dict[field_name], field_name, csv_column_types)
                row_in_sqlite_form.append(item_dict[field_name])
            else:
                if field_name == 'Row_Last_Updated':
                    row_in_sqlite_form.append(etld_lib_datetime.get_utc_datetime_sqlite_database_format())
                elif field_name == 'BATCH_DATE':
                    row_in_sqlite_form.append(batch_date)
                elif field_name == 'BATCH_NUMBER':
                    row_in_sqlite_form.append(batch_number)
                else:
                    row_in_sqlite_form.append("")  # Ensure blank is added to each required empty field

        return row_in_sqlite_form

    def prepare_database_field_vmpc(self, field_data: dict, field_name_tmp, csv_column_types: list):
        if field_name_tmp in csv_column_types:
            field_column_type = csv_column_types[field_name_tmp]
        else:
            field_column_type = ''
        if field_column_type == 'INTEGER':
            field_data = str(field_data).strip()
            if str(field_data).isnumeric():
                field_data = str(field_data)
            else:
                field_data = '0'
        elif field_data is None:
            field_data = ""
        elif 'DATE' in field_name_tmp:
            field_data = str(field_data).replace("T", " ").replace("Z", "")
            field_data = re.sub("\\..*$", "", field_data)
        elif isinstance(field_data, int):
            field_data = str(field_data)
        elif 'CVE_LIST' in field_name_tmp:
            if 'CVE' in field_data.keys():
                if isinstance(field_data['CVE'], dict):
                    one_item_dict = [field_data['CVE']]
                    field_data['CVE'] = one_item_dict
                    field_data = json.dumps(field_data)
                else:
                    field_data = json.dumps(field_data)
        elif not isinstance(field_data, str):
            field_data = json.dumps(field_data)

        return field_data


    def prepare_database_row_was(self, item_dict: dict, csv_columns: list, csv_column_types: list,
                                 batch_name: str, batch_date: str, batch_number: str):
        row_in_sqlite_form = []
        for field_name in csv_columns:  # Iterate through expected columns (contract)
            if field_name in item_dict.keys():  # Iterate through columns found in dictionary
                item_dict[field_name] = \
                    self.prepare_database_field_was(item_dict[field_name], field_name, csv_column_types)
                row_in_sqlite_form.append(item_dict[field_name])
            else:
                if field_name == 'Row_Last_Updated':
                    row_in_sqlite_form.append(etld_lib_datetime.get_utc_datetime_sqlite_database_format())
                elif field_name == 'BATCH_DATE':
                    row_in_sqlite_form.append(batch_date)
                elif field_name == 'BATCH_NUMBER':
                    row_in_sqlite_form.append(batch_number)
                else:
                    row_in_sqlite_form.append("")  # Ensure blank is added to each required empty field

        return row_in_sqlite_form


    def prepare_database_field_was(self, field_data: dict, field_name_tmp, csv_column_types: list):
        if field_name_tmp in csv_column_types:
            field_column_type = csv_column_types[field_name_tmp]
        else:
            field_column_type = ''
        if field_column_type == 'INTEGER':
            field_data = str(field_data).strip()
            if str(field_data).isnumeric():
                field_data = str(field_data)
            else:
                field_data = '0'
        elif field_data is None:
            field_data = ""
        elif 'lastBoot' in field_name_tmp:
            field_data = str(field_data).replace("T", " ").replace("Z", "")
            field_data = re.sub("\\..*$", "", field_data)
        elif 'Date' in field_name_tmp:
            field_data = str(field_data).replace("T", " ").replace("Z", "")
            field_data = re.sub("\\..*$", "", field_data)
        elif isinstance(field_data, int):
            field_data = str(field_data)
        elif not isinstance(field_data, str):
            field_data = json.dumps(field_data)

        return field_data

    # ADDED PCRS PREPARE ROW AND FIELD.

    def prepare_database_row_pcrs(self, item_dict: dict, csv_columns: list, csv_column_types: list,
                                 batch_name: str, batch_date: str, batch_number: str):
        row_in_sqlite_form = []
        for field_name in csv_columns:  # Iterate through expected columns (contract)
            if field_name in item_dict.keys():  # Iterate through columns found in dictionary
                item_dict[field_name] = \
                    self.prepare_database_field_pcrs(item_dict[field_name], field_name, csv_column_types)
                row_in_sqlite_form.append(item_dict[field_name])
            else:
                if field_name == 'Row_Last_Updated':
                    row_in_sqlite_form.append(etld_lib_datetime.get_utc_datetime_sqlite_database_format())
                elif field_name == 'BATCH_DATE':
                    row_in_sqlite_form.append(batch_date)
                elif field_name == 'BATCH_NUMBER':
                    row_in_sqlite_form.append(batch_number)
                else:
                    row_in_sqlite_form.append("")  # Ensure blank is added to each required empty field

        return row_in_sqlite_form


    def prepare_database_field_pcrs(self, field_data: dict, field_name_tmp, csv_column_types: list):
        if field_name_tmp in csv_column_types:
            field_column_type = csv_column_types[field_name_tmp]
        else:
            field_column_type = ''
        if field_column_type == 'INTEGER':
            field_data = str(field_data).strip()
            if str(field_data).isnumeric():
                field_data = str(field_data)
            else:
                field_data = '0'
        elif field_data is None:
            field_data = ""
        elif 'lastBoot' in field_name_tmp:
            field_data = str(field_data).replace("T", " ").replace("Z", "")
            field_data = re.sub("\\..*$", "", field_data)
        elif 'Date' in field_name_tmp:
            field_data = str(field_data).replace("T", " ").replace("Z", "")
            field_data = re.sub("\\..*$", "", field_data)
        elif 'created' in field_name_tmp:
            field_data = str(field_data).replace("T", " ").replace("Z", "")
            field_data = re.sub("\\..*$", "", field_data)
        elif isinstance(field_data, int):
            field_data = str(field_data)
        elif not isinstance(field_data, str):
            field_data = json.dumps(field_data)

        return field_data

    def prepare_database_row_asset_inventory(self, item_dict: dict, csv_columns: list, csv_column_types: list,
                                             batch_name: str, batch_date: str, batch_number: str):
        row_in_sqlite_form = []
        for field_name in csv_columns:  # Iterate through expected columns (contract)
            if field_name in item_dict.keys():  # Iterate through columns found in dictionary
                item_dict[field_name] = \
                    self.prepare_database_field_asset_inventory(item_dict[field_name], field_name, csv_column_types)
                row_in_sqlite_form.append(item_dict[field_name])
            else:
                if field_name == 'Row_Last_Updated':
                    row_in_sqlite_form.append(etld_lib_datetime.get_utc_datetime_sqlite_database_format())
                elif field_name == 'BATCH_DATE':
                    row_in_sqlite_form.append(batch_date)
                elif field_name == 'BATCH_NUMBER':
                    row_in_sqlite_form.append(batch_number)
                else:
                    row_in_sqlite_form.append("")  # Ensure blank is added to each required empty field

        return row_in_sqlite_form

    def prepare_database_field_asset_inventory(self, field_data: dict, field_name_tmp, csv_column_types: list):
        if field_name_tmp in csv_column_types:
            field_column_type = csv_column_types[field_name_tmp]
        else:
            field_column_type = ''
        if field_column_type == 'INTEGER':
            field_data = str(field_data).strip()
            if str(field_data).isnumeric():
                field_data = str(field_data)
            else:
                field_data = '0'
        elif field_data is None:
            field_data = ""
        elif 'lastBoot' in field_name_tmp:
            field_data = str(field_data).replace("T", " ").replace("Z", "")
            field_data = re.sub("\\..*$", "", field_data)
        elif 'Date' in field_name_tmp:
            field_data = str(field_data).replace("T", " ").replace("Z", "")
            field_data = re.sub("\\..*$", "", field_data)
        elif isinstance(field_data, int):
            field_data = str(field_data)
        elif not isinstance(field_data, str):
            field_data = json.dumps(field_data)

        return field_data

    # def prepare_database_field(self, field_data, field_name_tmp, field_column_type) -> dict:
    #     if field_column_type == 'INTEGER':
    #         field_data = str(field_data).strip()
    #         if str(field_data).isnumeric():
    #             field_data = str(field_data)
    #         else:
    #             field_data = '0'
    #     elif field_data is None:
    #         field_data = ""
    #     elif isinstance(field_data, str):
    #         if 'date' in str(field_name_tmp).lower():
    #             field_data = field_data.replace("T", " ").replace("Z", "")
    #             field_data = re.sub("\\..*$", "", field_data)
    #         elif 'dns' in str(field_name_tmp).lower():
    #             field_data = field_data.replace('\n', '')
    #     elif isinstance(field_data, int):
    #         field_data = str(field_data)
    #     elif not isinstance(field_data, str):
    #         field_data = json.dumps(field_data)
    #
    #     return field_data

    # def prepare_database_row_integer_fields(self, item: dict, database_columns: list, database_column_types: dict) -> list:
    #     for field_name in database_columns:
    #         if field_name in item.keys():
    #             field_column_type = ""
    #             if field_name in database_column_types:
    #                 if database_column_types[field_name] == 'INTEGER':
    #                     if item[field_name] =
    #                     item[field_name]
    #                 field_column_type = database_column_types[field_name]
    #
    #             item[field_name] = \
    #                 self.prepare_database_field(item[field_name], field_name, field_column_type)
    #             row_in_sqlite_form.append(item[field_name])
    #         else:
    #             row_in_sqlite_form.append("")  # Ensure blank is added to each required empty field
    #

    # def prepare_database_row(self, item: dict, database_columns: list, database_column_types: dict) -> list:
    #     row_in_sqlite_form = []
    #     for field_name in database_columns:
    #         if field_name in item.keys():
    #             field_column_type = ""
    #             if field_name in database_column_types:
    #                 field_column_type = database_column_types[field_name]
    #             item[field_name] = \
    #                 self.prepare_database_field(item[field_name], field_name, field_column_type)
    #             row_in_sqlite_form.append(item[field_name])
    #         else:
    #             row_in_sqlite_form.append("")  # Ensure blank is added to each required empty field
    #     return row_in_sqlite_form

    # def prepare_integer_columns_for_database(self, table_name: str, row: list):
    #     if table_name in self.pragma_dict:
    #         pragma = self.pragma_dict['table_name']
    #     else:
    #         pragma = self.get_pragma(self.cursor, table_name)
    #         self.pragma_dict[table_name] = pragma
    #
    #     row_fields = []
    #     for idx, fieldname in pragma:
    #         field_data = row[idx]
    #         if pragma[idx][2] == 'INTEGER':
    #             field_data = str(row[idx]).strip()
    #             if not str(row[idx]).isnumeric():
    #                 field_data = '0'
    #         row_fields.append(field_data)
    #     return row_fields

    def insert_or_replace_row(self, table_name: str, row: list):
        result = True
        try:
            self.cursor.execute(f"INSERT OR REPLACE INTO {table_name} "
                                f"VALUES({self.get_values_var(row)},datetime('now'))",
                                row)
        except Exception as e:
            etld_lib_functions.logger.error(
                f"Error inserting or replacing row into table: {table_name}, please retry after fixing error")
            etld_lib_functions.logger.error(f"Exception: {e}")
            result = False

        return result

    def insert_or_replace_row_pristine(self, table_name: str, row: list):
        result = True
        try:
            self.cursor.execute(f"INSERT OR REPLACE INTO {table_name} "
                                f"VALUES({self.get_values_var(row)})",
                                row)
        except Exception as e:
            etld_lib_functions.logger.error(
                f"Error inserting or replacing pristine row into table: {table_name}, please retry after fixing error")
            etld_lib_functions.logger.error(f"Exception: {e}")
            result = False

        return result

    def insert_or_replace_row_including_datetime(self, table_name: str, row: list, datetime: str):
        result = True
        try:
            row.append(datetime)
            self.cursor.execute(f"INSERT OR REPLACE INTO {table_name} "
                                f"VALUES({self.get_values_var(row)})",
                                row)
        except Exception as e:
            etld_lib_functions.logger.error(
                f"Error inserting or replacing row into table: {table_name}, please retry after fixing error")
            etld_lib_functions.logger.error(f"Exception: {e}")
            result = False

        return result

    # def insert_row_into_csv(self, table_name: str, row: list, csv_obj: etld_lib_sqlite_to_csv.SqliteTabletoCsvFile):
    #     result = True
    #     try:
    #         csv_obj.write_row(row)
    #     except Exception as e:
    #         etld_lib_functions.logger.error(
    #             f"Error writing csv record from {table_name}, please retry after fixing error")
    #         etld_lib_functions.logger.error(f"Exception: {e}")
    #         exit(1)
    #     return result

    def insert_unique_row_ignore_duplicates(self, table_name: str, row: list):
        result = True
        try:
            #self.cursor.execute(f"INSERT INTO {table_name} VALUES({self.get_values_var(row)},datetime('now'))", row)
            self.cursor.execute(f"INSERT INTO {table_name} VALUES({self.get_values_var(row)})", row)
        except sqlite3.IntegrityError as err:
            if 'unique constraint failed' in str(err).lower():
                result = "duplicate"
                #etld_lib_functions.logger.warning(
                #    f"Duplicate Key, inserting into: {table_name}, ignore, first field: {row[0]}")
                #etld_lib_functions.logger.error(f"Exception: {err}")
                pass
            else:
                etld_lib_functions.logger.error(
                    f"Integrity Error inserting into: {table_name}, please retry after fixing error")
                etld_lib_functions.logger.error(f"Exception: {err}")
                exit(1)

        except sqlite3.Error as err:
            etld_lib_functions.logger.error(
                f"Error inserting row into table: {table_name}, please retry after fixing error")
            etld_lib_functions.logger.error(f"Exception: {err}")
            exit(1)
        except Exception as e:
            etld_lib_functions.logger.error(
                f"Error inserting row into table: {table_name}, please retry after fixing error")
            etld_lib_functions.logger.error(f"Exception: {e}")
            exit(1)
        return result

    # def insert_unique_row(self, table_name: str, values_var: list, row: list, conflict: str):
    #     try:
    #         self.cursor.execute(f"INSERT INTO {table_name} VALUES({values_var},datetime('now'))", row)
    #     except sqlite3.Error as err:
    #         etld_lib_functions.logger.error(
    #             f"Error inserting row into table: {table_name}, please retry after fixing error")
    #         etld_lib_functions.logger.error(f"Exception: {err}")
    #         exit(1)
    #     except Exception as e:
    #         etld_lib_functions.logger.error(
    #             f"Error inserting row into table: {table_name}, please retry after fixing error")
    #         etld_lib_functions.logger.error(f"Exception: {e}")
    #         exit(1)

    def commit_changes(self):
        try:
            self.connection.commit()
        except Exception as e:
            etld_lib_functions.logger.error(
                f"Error committing to: {self.sqlite_file}, please retry after fixing error")
            etld_lib_functions.logger.error(f"Exception: {e}")
            exit(1)

    def close_connection(self):
        try:
            self.connection.close()
        except Exception as e:
            etld_lib_functions.logger.error(
                f"Error closing: {self.sqlite_file}, please retry after fixing error")
            etld_lib_functions.logger.error(f"Exception: {e}")
            exit(1)

    def vacuum_database(self):
        try:
            self.isolation_level = self.connection.isolation_level
            self.connection.isolation_level = None
            self.connection.execute("VACUUM")
            self.connection.isolation_level = self.isolation_level
        except Exception as e:
            etld_lib_functions.logger.error(
                f"Error vacuuming database: {self.sqlite_file}, please retry after fixing error")
            etld_lib_functions.logger.error(f"Exception: {e}")
            exit(1)

    def select_all_from_table(self, table_name):
        self.cursor.execute(f"SELECT * FROM {table_name}")
        return self.cursor.fetchall()

    def select_all_from_table_where_table_column_equals_value(
            self, table_name, table_column, where_table_column_equals_value):
        self.cursor.execute(f"SELECT * FROM {table_name} WHERE {table_column} LIKE '{where_table_column_equals_value}'")
        return self.cursor.fetchall()

    def select_from_table(self, select_statement):
        self.cursor.execute(f"{select_statement}")
        return self.cursor.fetchall()

    def select_all_from_status_table_where_status_name_equals_table_name(
            self, status_table_name, table_name):
        self.cursor.execute(f"SELECT * FROM {status_table_name}")
        return self.cursor.fetchall()

    @staticmethod
    def get_values_var(csv_columns: list):
        values_var = ""
        for column in csv_columns:
            values_var = f"{values_var}?, "
        values_var = re.sub("\?, $", "?", values_var)
        return values_var

    @staticmethod
    def status_table_prepare_fields_for_update(status_name='HOST_LIST_LOAD_STATUS', batch_date="", batch_number="0",
                                               status='COMPLETED', total_rows_added=0):
        status_name = status_name
        try:
            logfile_run_datetime = etld_lib_functions.logger_datetime
            logfile_workflow_name = etld_lib_functions.my_logger_program_name_for_database_routine
            status_detail_dict = {'BATCH_DATE': batch_date,
                                  'BATCH_NUMBER': batch_number,
                                  'STATUS': status,
                                  'TOTAL_ROWS_ADDED': total_rows_added,
                                  'LOG_WORKFLOW_DATE': logfile_run_datetime,
                                  'LOG_WORKFLOW_NAME': logfile_workflow_name
                                  }
        except AttributeError:
            status_detail_dict = {'BATCH_DATE': batch_date,
                                  'BATCH_NUMBER': batch_number,
                                  'STATUS': status,
                                  'TOTAL_ROWS_ADDED': total_rows_added,
                                  'LOG_WORKFLOW_DATE': "",
                                  'LOG_WORKFLOW_NAME': ""
                                  }

        # change to support text field of batch name - status_count (last_batch_processed) = int(batch_number)
        last_batch_processed = batch_number
        return status_name, status_detail_dict, last_batch_processed


    @staticmethod
    def status_table_prepare_general_data_for_update(
            status_name='LOG_DATA', status_detail_dict={}, last_batch_processed="FALSE"):
            # LOG_DATA - status_detail_dict will be all logging relevant for job.
            #          - last_batch_prcoessed will be set to FALSE, ERROR, TRUE
            #          - If last_batch_processed is set to FALSE or ERROR, do not move forward with ETL.

        return status_name, status_detail_dict, last_batch_processed

    def update_status_table_general_data_update(self, status_table_name,
                                                status_name, status_detail_dict, last_batch_processed):

        self.status_table_insert_or_replace_row_in_sqlite(
            table_name=status_table_name,
            status_name=status_name,
            status_detail_dict=status_detail_dict,
            last_batch_processed=last_batch_processed
        )
        self.status_table_select_status_row_from_sqlite(
            table_name=status_table_name,
            table_column='STATUS_NAME',
            where_table_column_equals_value=status_name)

        etld_lib_functions.logger.info(f"Updated {status_table_name} - "
                                       f"status_name={status_name} - "
                                       f"last_batch_processed={last_batch_processed}")

    def get_database_status_detail(self, status_name='ALL_TABLES_LOADED_SUCCESSFULLY'):
        database_table_names_list = self.get_all_table_names_from_database()
        status_detail = {}
        for database_table_name in database_table_names_list:
            if str(database_table_name).endswith("_Status"):
                self.cursor.execute(f"select * from {database_table_name} where STATUS_NAME == '{status_name}'")
                row = self.cursor.fetchall()
                if row is None:
                    etld_lib_functions.logger.info(f"TABLE: {database_table_name} "
                                                   f"did not contain STATUS_NAME:{status_name} "
                                                   f"required to retrieve STATUS_DETAIL")
                    status_detail = {}
                if 'STATUS_DETAIL' in row.keys():
                    json_string = row['STATUS_DETAIL']
                    status_detail = json.loads(json_string)
        return status_detail

    def status_table_insert_or_replace_row_in_sqlite(self,
                                                     table_name: str, status_name: str, status_detail_dict: dict,
                                                     last_batch_processed: str, commit_changes: bool = True):
        status_detail_json = json.dumps(status_detail_dict)
        row_last_updated_datetime = etld_lib_datetime.get_utc_datetime_sqlite_database_format()
        row_to_insert = [status_name,
                         status_detail_json,
                         last_batch_processed,
                         row_last_updated_datetime
                         ]
        result = self.insert_or_replace_row_pristine(
            table_name=table_name,
            row=row_to_insert
        )
        if commit_changes is True:
            self.commit_changes()

    def status_table_select_status_row_from_sqlite(self,
                                                   table_name: str, table_column: str,
                                                   where_table_column_equals_value: str):
        result = self.select_all_from_table_where_table_column_equals_value(
            table_name=table_name,
            table_column=table_column,
            where_table_column_equals_value=where_table_column_equals_value
        )
        status_name = result[0]['STATUS_NAME']
        status_detail_dict = json.loads(result[0]['STATUS_DETAIL'])
        last_batch_processed = result[0]['LAST_BATCH_PROCESSED']
        return status_name, status_detail_dict, last_batch_processed

    def update_status_table(self,
                            batch_date: str,
                            batch_number: str,
                            total_rows_added_to_database: int,
                            status_table_name: str,
                            status_table_columns: list,
                            status_table_column_types: list,
                            status_name_column: str = 'TABLE_LOAD_STATUS',
                            status_column: str = 'begin'):

        status_column_dict = {'begin': 'BEGIN LOADING BATCH',
                              'end': 'END LOADING BATCH',
                              'final': 'FINAL BATCH LOADED',
                              }

        if status_column in status_column_dict.keys():
            status_column_display = status_column_dict[status_column]
        else:
            status_column_display = status_column

        status_name, status_detail_dict, last_batch_processed = \
            self.status_table_prepare_fields_for_update(
                status_name=status_name_column,
                batch_date=batch_date,
                batch_number=batch_number,
                status=status_column_display,
                total_rows_added=total_rows_added_to_database)

        self.status_table_insert_or_replace_row_in_sqlite(
            table_name=status_table_name,
            status_name=status_name,
            status_detail_dict=status_detail_dict,
            last_batch_processed=str(last_batch_processed)
        )

        status_name, status_detail_dict, last_batch_processed = \
            self.status_table_select_status_row_from_sqlite(
                table_name=status_table_name,
                table_column='STATUS_NAME',
                where_table_column_equals_value=status_name)
        etld_lib_functions.logger.info(f"Updated {status_table_name} - "
                                       f"status_name={status_name} - status_detail_dict={status_detail_dict}")

        return status_name, status_detail_dict, last_batch_processed

    def bulk_insert_csv_file(self, table_name, csv_file_name, csv_columns, message="",
                             display_progress_counter_at_this_number=10000,
                             delimiter=',', header_present=True,
                             key_column_index="", key_to_look_for_in_data=""):
        csv.field_size_limit(sys.maxsize)
        sqlite_rows_written = \
            etld_lib_functions.DisplayCounterToLog(
                display_counter_at=display_progress_counter_at_this_number,
                logger_func=etld_lib_functions.logger.info,
                display_counter_log_message=f"count {table_name} sqlite rows written")
        # etld_lib_functions.logger.info(f"csv.field_size_limit(sys.maxsize): {str(sys.maxsize)}")
        try:
            values_var = ""
            for column in csv_columns:
                values_var = f"{values_var}?, "
            values_var = re.sub("\?, $", "?", values_var)
            csv_headers = ""
            csv_row_list = []
            bulk_insert_count = 100000
            with open(csv_file_name, "rt", encoding='utf-8') as csv_file:
                csv_reader = csv.reader(csv_file, delimiter=delimiter)
                self.count_rows_added_to_table = 0
                if header_present is True:
                    csv_headers = next(csv_reader)  # Skip Header
                for row in csv_reader:
                    if isinstance(key_column_index, int):
                        if key_column_index <= len(row):
                            if key_to_look_for_in_data in row[key_column_index]:
                                pass  # Insert row with key.  Ex. 'etl_was 20220122180000'
                            else:
                                continue
                    self.cursor.execute(
                        f"INSERT OR REPLACE INTO {table_name} VALUES({values_var},datetime('now'))", row)
                    sqlite_rows_written.update_counter_and_display_to_log()
                    self.count_rows_added_to_table += 1
                self.commit_changes()
                sqlite_rows_written.display_final_counter_to_log()

        except sqlite3.Error as err:
            etld_lib_functions.logger.error(
                f"Error inserting row into table: {table_name}, please retry after fixing error")
            etld_lib_functions.logger.error(f"Exception: {err}")
            exit(1)
        except Exception as e:
            etld_lib_functions.logger.error(f"Error creating table from: {csv_file_name}, retry after fixing error")
            etld_lib_functions.logger.error(f"Row Number: {self.count_rows_added_to_table}")
            etld_lib_functions.logger.error(f"Exception: {e}")
            exit(1)



#!/usr/bin/env python3
import sys
import timeit
from pathlib import Path
from qualys_etl.etld_lib import etld_lib_functions
from qualys_etl.etld_lib import etld_lib_config
#from qualys_etl.etld_lib import etld_lib_credentials
from qualys_etl.etld_lib import etld_lib_authentication_objects

from qualys_etl.etld_host_list import host_list_03_extract_controller
from qualys_etl.etld_host_list import host_list_06_distribution

global start_time
global stop_time


def host_list_03_extract_controller_wrapper(module_function=host_list_03_extract_controller, message=""):
    etld_lib_functions.logger.info(f"start {module_function} {message}")
    module_function.main()
    etld_lib_functions.logger.info(f"end   {module_function}")


def host_list_06_distribution_wrapper(
        module_function=host_list_06_distribution, message=""):
    etld_lib_functions.logger.info(f"start {module_function} {message}")
    module_function.main()
    etld_lib_functions.logger.info(f"end   {module_function}")


def begin_message_info():
    etld_lib_functions.logger.info(f"data directory: {etld_lib_config.qetl_user_data_dir}")


def begin_host_list_02_workflow_manager():
    global start_time
    start_time = timeit.default_timer()
    etld_lib_functions.logger.info(f"__start__ host_list_etl_workflow {str(sys.argv)}")
    begin_message_info()

# def host_list_start_wrapper():
#     global start_time
#     start_time = timeit.default_timer()
#     etld_lib_functions.logger.info(f"__start__ host_list_etl_workflow {str(sys.argv)}")
#     etld_lib_functions.logger.info(f"data directory: {etld_lib_config.qetl_user_data_dir}")
#     etld_lib_functions.logger.info(f"config file:    {etld_lib_config.qetl_user_config_settings_yaml_file}")
#     etld_lib_functions.logger.info(f"cred yaml file: {etld_lib_credentials.cred_file}")
#     etld_lib_functions.logger.info(f"cookie file:    {etld_lib_credentials.cookie_file}")

def end_host_list_02_workflow_manager():
    global start_time
    global stop_time

    stop_time = timeit.default_timer()
    etld_lib_functions.logger.info(f"runtime for host_list_etl_workflow in seconds: {stop_time - start_time:,}")
    etld_lib_functions.logger.info(f"__end__ host_list_etl_workflow {str(sys.argv)}")

# def host_list_end_wrapper():
#     global start_time
#     global stop_time
#
#     stop_time = timeit.default_timer()
#     etld_lib_functions.logger.info(f"runtime for host_list_etl_workflow in seconds: {stop_time - start_time:,}")
#     etld_lib_functions.logger.info(f"__end__ host_list_etl_workflow {str(sys.argv)}")


def host_list_etl_workflow():
    try:
        begin_host_list_02_workflow_manager()
        host_list_03_extract_controller_wrapper(
            message=f"vm_processed_after={etld_lib_config.host_list_vm_processed_after}")
        host_list_06_distribution_wrapper()
        end_host_list_02_workflow_manager()
    except Exception as e:
        etld_lib_functions.logger.error(f"Error occurred, please investigate {sys.argv}")
        etld_lib_functions.logger.error(f"Exception: {e}")
        exit(1)


def main():
    host_list_etl_workflow()


if __name__ == "__main__":
    etld_lib_functions.main(my_logger_prog_name='host_list_02_workflow_manager')
    etld_lib_config.main()
    #etld_lib_credentials.main()
    etld_lib_authentication_objects.main()
    main()



#!/usr/bin/env python3
import sys
import timeit
import time
import fcntl
from pathlib import Path
import multiprocessing
from qualys_etl.etld_lib import etld_lib_functions
from qualys_etl.etld_lib import etld_lib_config
#from qualys_etl.etld_lib import etld_lib_credentials
from qualys_etl.etld_lib import etld_lib_authentication_objects

from qualys_etl.etld_host_list_detection import host_list_detection_03_extract_controller
from qualys_etl.etld_host_list_detection import host_list_detection_05_transform_load_xml_to_sqlite
from qualys_etl.etld_host_list_detection import host_list_detection_06_distribution
from qualys_etl.etld_host_list import host_list_02_workflow_manager
from qualys_etl.etld_knowledgebase import knowledgebase_02_workflow_manager

global start_time
global stop_time


def host_list_02_workflow_manager_wrapper(module_function=host_list_02_workflow_manager, message=""):
    etld_lib_functions.logger.info(f"start {module_function} {message}")
    module_function.main()
    etld_lib_functions.logger.info(f"end   {module_function}")


def knowledgebase_02_workflow_manager_wrapper(
       module_function=knowledgebase_02_workflow_manager, message=""
):
    etld_lib_functions.logger.info(f"start {module_function} {message}")
    module_function.main()
    etld_lib_functions.logger.info(f"end   {module_function}")


def host_list_detection_03_extract_controller_wrapper(module_function=host_list_detection_03_extract_controller, message=""):
    etld_lib_functions.logger.info(f"start {module_function} {message}")
    module_function.main()
    etld_lib_functions.logger.info(f"end   {module_function}")


def host_list_detection_05_transform_load_xml_to_sqlite_wrapper(
        module_function=host_list_detection_05_transform_load_xml_to_sqlite):
    etld_lib_functions.logger.info(f"start {module_function}")
    module_function.main()
    etld_lib_functions.logger.info(f"end   {module_function}")


def host_list_detection_06_distribution_wrapper(module_function=host_list_detection_06_distribution):
    etld_lib_functions.logger.info(f"start {module_function}")
    module_function.main()
    etld_lib_functions.logger.info(f"end   {module_function}")


def begin_message_info():
    etld_lib_functions.logger.info(f"data directory: {etld_lib_config.qetl_user_data_dir}")
    etld_lib_functions.logger.info(f"config file:    {etld_lib_config.qetl_user_config_settings_yaml_file}")


def begin_host_list_detection_02_workflow_manager():
    global start_time
    start_time = timeit.default_timer()
    etld_lib_functions.logger.info(f"__start__ host_list_detection_etl_workflow {str(sys.argv)}")
    begin_message_info()


def end_host_list_detection_02_workflow_manager():
    global start_time
    global stop_time

    stop_time = timeit.default_timer()
    etld_lib_functions.logger.info(f"runtime for host_list_detection_etl_workflow in seconds: {stop_time - start_time:,}")
    etld_lib_functions.logger.info(f"__end__ host_list_detection_etl_workflow {str(sys.argv)}")


def spawn_knowledgebase_02_workflow_manager_wrapper():
    try:
        etld_lib_config.kb_last_modified_after = ""
        process = \
            multiprocessing.Process(target=knowledgebase_02_workflow_manager_wrapper,
                                    name='knowledgebase_02_workflow_manager',
                                    args=(knowledgebase_02_workflow_manager, ""))
        return process
    except Exception as e:
        etld_lib_functions.logger.error(f"Error spawn_knowledgebase_02_workflow_manager_wrapper")
        etld_lib_functions.logger.error(f"Exception: {e}")
        raise Exception("Raise Exception from spawn_knowledgebase_02_workflow_manager_wrapper")


def get_host_list_data():

    try:
        etld_lib_config.host_list_detection_vm_processed_after = etld_lib_config.host_list_vm_processed_after
        long_message = f"host_list_detection_vm_processed_after = " \
                       f"host_list_vm_processed_after={etld_lib_config.host_list_vm_processed_after}"
        host_list_02_workflow_manager_wrapper(message=long_message)
    except Exception as e:
        etld_lib_functions.logger.error(f"Error get_host_list_data")
        etld_lib_functions.logger.error(f"Exception: {e}")
        raise Exception("Raise Exception from host_list_02_workflow_manager_wrapper")


def get_knowledgebase_and_host_list_data():

    try:
        knowledgebase_process = spawn_knowledgebase_02_workflow_manager_wrapper()

        proc_list = [
            knowledgebase_process,
        ]

        knowledgebase_process.daemon = True
        knowledgebase_process.start()
        get_host_list_data()  # blocking as host list database is built.

        proc_timeout = 3600  # 1 hour catch all for knowledgebase
        for proc_item in proc_list:
            proc_item.join(timeout=proc_timeout)

        for proc_item in proc_list:
            if proc_item.exitcode == 0:
                etld_lib_functions.logger.info(f"{proc_item.name} exited successfully, exit code: {proc_item.exitcode}")
            else:
                etld_lib_functions.logger.warning(
                    f"{proc_item.name} exited after .join timeout of {proc_timeout} seconds "
                    f"with error code: {proc_item.exitcode}")
                etld_lib_functions.logger.warning(
                    f"{proc_item.name} failed, please investigate why job ran so long.")
                for proc_item_to_kill in proc_list:
                    proc_item_to_kill.kill()
                raise Exception("get_knowledgebase_and_host_list_data multiprocessing exception")

    except Exception as e:
        etld_lib_functions.logger.warning(f"Exception: {e}")
        raise Exception("get_knowledgebase_and_host_list_data outer catch exception")


def get_knowledgebase_and_host_list_controller():
    retry_attempts = 6
    retry_attempt = 0
    retry_sleep_time = 300
    while retry_attempt < retry_attempts:
        retry_attempt += 1
        try:
            with open(etld_lib_config.kb_lock_file, 'wb+') as lock_knowledgebase_fcntl, \
                    open(etld_lib_config.host_list_lock_file, 'wb+') as lock_host_list_fcntl:        # If locked, exit.
                fcntl.flock(lock_knowledgebase_fcntl, fcntl.LOCK_EX | fcntl.LOCK_NB)
                fcntl.flock(lock_host_list_fcntl, fcntl.LOCK_EX | fcntl.LOCK_NB)
                get_knowledgebase_and_host_list_data()
                break # Success
        except Exception as e:
            etld_lib_functions.logger.warning(f"KnowledgeBase or Host List Program Exception.  {__file__} ")
            etld_lib_functions.logger.warning(f"Retrying in {retry_sleep_time} seconds")
            time.sleep(retry_sleep_time)
    else:
        etld_lib_functions.logger.error(f"Retried {retry_attempt} times with no success.")
        etld_lib_functions.logger.error(f"Check to see if KnowledgeBase or Host List Program is already running.  {__file__} ")
        exit(1)


def host_list_detection_etl_workflow():
    begin_host_list_detection_02_workflow_manager()
    get_knowledgebase_and_host_list_controller()

    try:
        host_list_detection_03_extract_controller_wrapper(
            message=f"vm_processed_after={etld_lib_config.host_list_detection_vm_processed_after}")
    except Exception as e:
        etld_lib_functions.logger.error(f"Error host_list_detection_03_extract_controller_wrapper, please investigate {sys.argv}")
        etld_lib_functions.logger.error(f"Exception: {e}")
        exit(1)

    try:
        host_list_detection_06_distribution_wrapper()
    except Exception as e:
        etld_lib_functions.logger.error(f"Error host_list_detection_06_distribution_wrapper, please investigate {sys.argv}")
        etld_lib_functions.logger.error(f"Exception: {e}")
        exit(1)

    end_host_list_detection_02_workflow_manager()


def main():
    host_list_detection_etl_workflow()


if __name__ == "__main__":
    etld_lib_functions.main(my_logger_prog_name='host_list_detection_02_workflow_manager')
    etld_lib_config.main()
    #etld_lib_credentials.main()
    etld_lib_authentication_objects.main()
    main()
